#include "parser.h"
#include "helpers.h"

Node parse_token(Compiler* self, Token token) {
    switch(token.type) {
    case TOKEN_TYPEIDENTIFIER:
        return parse_type(self, token.value);
    case TOKEN_IDENTIFIER:
        return parse_identifier(self, token.value);
    case TOKEN_NUMBER:
        return parse_number(self, token.value);
    default:
        syntax_error("unexpected token \"%s\"", token.value);
    }
}

Nodes parse_until(Compiler* self, TokenType end_token) {
    Nodes collected = { 0 };

    Token token;
    while(next(self, &token)) {
        if(token.type == end_token) {
            break;
        }

        Nodes_push(&collected, parse_token(self, token));
    }

    return collected;
}

Node parse_type(Compiler* self, const char* type) {
    printf("parse type %s\n", type);
    Token token;
    next(self, &token);

    switch(token.type) {
    case TOKEN_IDENTIFIER:
        return parse_type_ident_pair(self, type, token.value);
        break;
    default:
        syntax_error("unexpected token \"%s\" after type \"%s\"", token.value, type);
    }
}

Node parse_identifier(Compiler* self, const char* name) {
    Token token;
    next(self, &token);

    switch(token.type) {
    case TOKEN_OPAREN:
        return parse_function_call(self, name);
        break;
    default:
        syntax_error("unexpected token \"%s\" after identifier \"%s\"", token.value, name);
    }
}

Node parse_number(Compiler* self, const char* value) {
    return (Node){
        .type = NODE_NUMBER,
        .number = atoll(value),
    };
}

Node parse_type_ident_pair(Compiler* self, const char* type, const char* name) {
    printf("parse type ident pair %s %s\n", type, name);
    Token token;
    next(self, &token);

    switch(token.type) {
    case TOKEN_OPAREN:
        return parse_function(self, type, name);
    default:
        syntax_error("unexpected token \"%s\" after \"%s %s\"", token.value, type, name);
    }
}

Node parse_function(Compiler* self, const char* type, const char* name) {
    printf("parse function %s %s\n", type, name);
    Nodes args = { 0 };
    while(true) {
        Token arg_type;
        next(self, &arg_type);
        if(arg_type.type == TOKEN_CPAREN) {
            break;
        }

        Token arg_name;
        next(self, &arg_name);

        Nodes_push(&args,
            (Node){
                .type = NODE_ARG_DECL,
                .arg_decl = { .type = arg_type.value, .name = arg_name.value },
            });

        Token token;
        next(self, &token);
        switch(token.type) {
        case TOKEN_COMMA:
            continue;
        case TOKEN_CPAREN:
            break;
        default:
            syntax_error("expected comma or closing parenthesis, found %s", token.value);
        }

        break;  // only reached if didnt continue or error
    }

    Token token;
    if(next(self, &token)) {
        switch(token.type) {
        case TOKEN_SEMICOLON:
            return parse_function_decl(self, type, name, args);
        case TOKEN_OBRACE:
            return parse_function_impl(self, type, name, args);
        default:
            syntax_error("expected semicolon or opening brace found, %s", token.value);
        }
    } else {
        syntax_error("expected token, found eof");
    }
}

Node parse_function_decl(Compiler* self, const char* type, const char* name, Nodes args) {
    return (Node){
        .type = NODE_FUNCTION_DECL,
        .function_decl = { .type = type, .name = name, .args = args },
    };
}

Node parse_function_impl(Compiler* self, const char* type, const char* name, Nodes args) {
    Nodes body = parse_until(self, TOKEN_CBRACE);

    return (Node){
        .type = NODE_FUNCTION_IMPL,
        .function_impl = { .type = type, .name = name, .args = args, .body = body },
    };
}

Node parse_function_call(Compiler* self, const char* name) {
    printf("parse function call %s\n", name);
    Nodes args = parse_until(self, TOKEN_CPAREN);

    parse_semicolon(self);
    return (Node){
        .type = NODE_FUNCTION_CALL,
        .function_call = { .name = name, .args = args },
    };
}

void parse_semicolon(Compiler* self) {
    Token token;
    next(self, &token);

    switch(token.type) {
    case TOKEN_SEMICOLON:
        return;
    default:
        syntax_error("expected semicolon, found \"%s\"", token.value);
    }
}
